use std::cmp::max;
use std::time::Duration;
use log::{info, debug, warn, error};
use proxmox_api::nodes::node::lxc::vmid::config::PutParams;
use proxmox_api::nodes::node::lxc::vmid::move_volume;
use proxmox_api::types::VmId;
use proxmox_api::UreqError;
use proxmox_api::UreqError::Ureq;
use sea_orm::*;
use sea_orm::ActiveValue::Set;
use serde::{Deserialize, Serialize};
use tokio::time::sleep;
use ureq::Error::Status;
use crate::api::cluster::carrier::{provision_carrier, terminate_carrier};
use crate::api::cluster::node::migrate_node;
use crate::api::entity::nodes::{lock_first_available, P5xError};
use crate::api::entity::{nodes, volumes};
use crate::api::services::{ssh_run_trimmed, Services};


/** Parameters required from an API call to manage a volume. */
#[derive(Serialize, Deserialize, FromForm)]
pub struct VolumeParams {
    pub id: Option<i32>,
    pub name: String,
    #[serde(default)]
    pub size_in_bytes: i64,
    pub mountpoint: Option<String>,
}

impl VolumeParams {
    /** Look up a volume by name and get its params. */
    pub async fn resolve(svc: &Services<'_>, name: &str) -> Result<Option<VolumeParams>, DbErr> {
        volumes::Entity::find()
            .filter(volumes::Column::Name.eq(name))
            .one(svc.db)
            .await
            .map(|v| v.map(|v| v.into()))
    }
}


/** Create a new PVE volume of the given size. */
pub async fn create(
    svc: &Services<'_>,
    name: &str,
    size_in_bytes: i64,
) -> Result<volumes::Model, P5xError> {
    info!("Creating volume {name} with size {}KiB", size_in_bytes / 1024);

    // Make sure the volume name is unique
    let existing = volumes::Entity::find()
        .filter(volumes::Column::Name.eq(name))
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    if let Some(_) = existing {
        return Err(P5xError::BadPrecondition("Cannot create volume: a volume with the specified name already exists"));
    }

    // Lock a node in the k8s cluster
    let reason = format!("Creating volume: {name}");
    let (node, _lock) = lock_first_available(svc.db, Some(&reason))
        .await.map_err(P5xError::LockErr)?;

    // Get the next available mountpoint ID
    let conf = node.config(svc)?;
    let mp_id = conf.next_nth("mp");
    info!("Volume {name} will become mp{mp_id} on node {} ({})", node.hostname, node.pve_id);

    // Generate a new mountpoint entry for the node's config
    let storage = &svc.config.pve_storage_pool;
    let size_in_gib = max((size_in_bytes as u64).div_ceil(1024 * 1024 * 1024) as i64, 1);
    let line = format!("{storage}:{size_in_gib},mp=/mnt/p5x-{name},backup=1");
    debug!("Volume {name}: {line}");

    let mut params = PutParams::default();
    params.mps.insert(mp_id, line);

    // Update the node config to create the volume
    debug!("Patching PVE config for volume {name}");
    let vm_id = VmId::new(i64::from(node.pve_id)).unwrap();
    let res = svc.pve_node(&node.pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .vmid(vm_id)
        .config()
        .put(params);

    // This is necessary because PUT returns {data: null} on success,
    // which the UreqClient flags as an unknown error.
    if let Err(UreqError::EncounteredErrors(e)) = res {
        return Err(P5xError::PveError(UreqError::EncounteredErrors(e)));
    }

    // Stupid hack is stupid, but we don't get back a UPID to wait on the vol to be created
    info!("Successfully patched PVE config. Waiting for volume {name} to appear");
    sleep(Duration::from_secs(5)).await;

    // Load the updated config
    debug!("Loading updated node config for volume {name}");
    let conf = node.config(svc)?;
    let mount = conf.get(&format!("mp{mp_id}"))
        .ok_or(P5xError::BadPrecondition("Could not find mountpoint in config after creating volume"))?;
    debug!("Found mountpoint details for volume {name}: {mount}");

    // Parse the disk name from the config
    // synology-scsi-lun:vm-103-disk-1,mp=/mnt/p5x-test0,backup=1,size=1G  // fixme: how does this behave for NFS?
    let name_offset = storage.len() + 1;  // 1 for the colon (:)
    let disk_name = mount[name_offset..].split(",").next().unwrap();

    // Persist the volume
    debug!("Inserting record into volumes table for volume {name} ({disk_name})");
    let res = volumes::ActiveModel {
        name: Set(name.to_string()),
        size_in_bytes: Set(size_in_bytes),
        pve_node_id: Set(node.pve_id),
        mountpoint: Set(Some(format!("/mnt/p5x-{name}"))),
        mountpoint_identifier: Set(Some(format!("mp{mp_id}"))),
        disk_name: Set(Some(disk_name.to_string())),
        ..Default::default()
    }
        .save(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    info!("Wrote new volume {name} to volumes table");
    let vol = res.try_into_model()
        .map_err(P5xError::DbErr)?;

    debug!("Volume {name} creation completed successfully");
    Ok(vol)
}


/** Mount a volume to the specified mountpoint on the LXC container it is attached to. */
pub async fn mount(
    svc: &Services<'_>,
    params: &VolumeParams,
) -> Result<(), P5xError> {
    info!("Mounting volume {}", params.name);
    if params.mountpoint.is_none() {
        return Err(P5xError::BadPrecondition("Missing required mountpoint parameter"));
    }
    let mountpoint = params.mountpoint.as_ref().unwrap();

    // Look up the volume instance
    let vol = volumes::resolve(svc, params).await?;

    // Make sure the volume isn't already mounted somewhere
    if vol.mountpoint.is_some() {
        return Err(P5xError::BadPrecondition("Tried to mount volume that is already mounted"));
    }

    // Lock the volume's node
    let node = vol.node(svc).await?;
    debug!("Locking node {} to mount volume {} at {}", node.hostname, params.name, mountpoint);
    let reason = format!("Mounting volume {} at {}", params.name, mountpoint);
    let _lock = node.lock(svc, Some(&reason));

    // Find the next available mountpoint identifier
    let qualified_name = vol.qualified_name(svc).await?;
    let mountpoint_identifier = node.config(svc)?.next_nth("mp");
    let mount_line = format!("{qualified_name},mp={mountpoint},backup=1");

    // Patch the node's config to mount the volume
    let pve_node = svc.pve_node(&node.pve_host)
        .map_err(P5xError::ServiceError)?;

    let mut put_params = PutParams::default();
    put_params.mps.insert(mountpoint_identifier, mount_line);

    debug!("Patching node config to mount volume {} ({put_params:?})", params.name);
    let res = pve_node.lxc()
        .vmid(node.vm_id())
        .config()
        .put(put_params);

    if let Err(Ureq(Status(_, ires))) = res {
        debug!("PVE response: {}", ires.into_string().unwrap());
    }

    // This is necessary because PUT returns {data: null} on success,
    // which the UreqClient flags as an unknown error.
    /*if let Err(UreqError::EncounteredErrors(e)) = res {
        return Err(P5xError::PveError(UreqError::EncounteredErrors(e)));
    }*/

    // Persist the volume
    debug!("Persisting mount changes to volume {} in database", params.name);
    let mut vol = vol.into_active_model();
    vol.mountpoint_identifier = Set(Some(format!("mp{mountpoint_identifier}")));
    vol.mountpoint = Set(Some(mountpoint.to_string()));
    vol.save(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    info!("Successfully mounted volume {} at {} on {}", params.name, mountpoint, node.hostname);
    Ok(())
}


/** Unmount a volume from the LXC container it is attached to. */
pub async fn unmount(
    svc: &Services<'_>,
    params: &VolumeParams,
) -> Result<(), P5xError> {
    info!("Unmounting volume {}", params.name);

    // Look up the volume instance
    let vol = volumes::resolve(svc, params).await?;

    // Check that the volume is actually mounted
    if vol.mountpoint.is_none() || vol.mountpoint_identifier.is_none() {
        warn!("Tried to unmount volume {}#{}, but volume is not mounted", params.name, vol.volume_id);
        return Ok(());
    }

    // Make sure we have the registered disk name
    if vol.disk_name.is_none() {
        return Err(P5xError::BadPrecondition("Tried to unmount volume without a disk_name set"));
    }

    // Unmount the disk's filesystem from within the K8s node
    debug!("Unmounting volume {} ({}) from K8s host", vol.name, vol.mountpoint.as_ref().unwrap());
    let cmd = format!("umount '{}'", vol.mountpoint.as_ref().unwrap());
    let node = vol.node(svc).await?;
    node.ssh_run_trimmed(svc, &cmd)?;

    // Unmount the disk's filesystem from the PVE node's shadow tree (see UNMOUNT_NOTES below)
    // -- Find the PID of the LXC container itself
    debug!("Attempting to identify host PID for node {}", &node.pve_host);
    let pve_ssh = svc.pve_ssh(&node.pve_host)
        .map_err(P5xError::ServiceError)?;
    let cmd = format!("lxc-info -n {} -p", node.pve_id);
    let ct_pid = ssh_run_trimmed(&pve_ssh, &cmd)?;
    let ct_pid = ct_pid
        .split("PID:")
        .nth(1)
        .ok_or(P5xError::BadPrecondition("Failed to parse PID of node's LXC container"))?
        .trim();

    // -- Find the parent PID (where the shadow tree exists)
    debug!("Attempting to identify parent PID for node {} (pid: {})", &node.pve_host, ct_pid);
    let cmd = format!("ps -o ppid= -p {ct_pid}");
    let parent_pid = ssh_run_trimmed(&pve_ssh, &cmd)?;

    // -- Unmount the disk's filesystem from the shadow tree's mount namespace:
    let mp_id = vol.mountpoint_identifier.clone().unwrap();
    debug!("Unmounting volume {} from shadow namespace (pid: {}) {}", &vol.name, parent_pid, &mp_id);
    let cmd = format!("nsenter --target {parent_pid} --mount /bin/bash -c 'umount /var/lib/lxc/.pve-staged-mounts/{}'", mp_id);
    ssh_run_trimmed(&pve_ssh, &cmd)?;

    // For LVM-type storage pools, we also need to deactivate the logical volume
    let pool_name = &svc.config.pve_storage_pool;
    let pool_driver = &svc.config.pve_storage_driver;
    if pool_driver == "lvm" {
        let cmd = format!("lvchange -aln '/dev/{pool_name}/{}'", vol.disk_name.as_ref().unwrap());
        ssh_run_trimmed(&pve_ssh, &cmd)?;
    }

    // Patch the PVE node's config to mark the volume as an unused disk
    let qualified_name = vol.qualified_name(svc).await?;
    let lock_reason = format!("Unmounting volume {}", vol.volume_id);
    let mut vol = vol.into_active_model();
    node.mutate_config(
        svc,
        Some(&lock_reason),
        |mut conf| {
            let next_unused = conf.next_nth("unused");
            conf.replace(
                |s| s.starts_with(&mp_id),
                |_| Some(format!("unused{next_unused}: {qualified_name}")),
            );

            vol.mountpoint_identifier = Set(Some(format!("unused{next_unused}")));

            conf
        }
    )
        .await?;

    // Persist the volume changes
    vol.mountpoint = Set(None);
    vol.save(svc.db).await
        .map_err(P5xError::DbErr)?;

    info!("Successfully unmounted volume {} from node {}", &params.name, &node.hostname);
    Ok(())
}

// UNMOUNT_NOTES
// Okay, here's some fucky-wucky shit:
// To avoid security vulnerabilities where hosts can umount their disks and break the
// firewall between CT and host, when a disk is mounted to a CT, Proxmox opens a clone
// of the mount in a special directory in the mount namespace of the container process'
// parent (i.e. the "monitor" process).
// If we umount the disk from the container, but not the monitor process, we won't be
// able to reattach any disks to the same mpX path (mp0, mp1, mp2, &c.)
// So to get around this, we (1) look up the container process, (2) figure out the
// monitor process, then (3) umount the clone.
// This was *such* a pain in the ass to figure out, but it's a testament to open-
// source that I was able to do it at all.


/** Delete the given volume from the PVE cluster. */
pub async fn delete(
    svc: &Services<'_>,
    params: &VolumeParams,
) -> Result<(), P5xError> {
    info!("Deleting volume {}", params.name);

    // Look up the existing volume
    let vol = volumes::Entity::find()
        .filter(volumes::Column::Name.eq(&params.name))
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?
        .ok_or(P5xError::BadPrecondition("Could not delete volume: unable to find volume with that name"))?;

    // Make sure we have a mountpoint identifier
    if vol.mountpoint_identifier.is_none() {
        return Err(P5xError::BadPrecondition("Could not delete volume: volume has no mountpoint identifier"));
    }

    // Patch the PVE config to delete the volume from the config
    let node = vol.node(svc).await?;
    debug!("Patching node {} config to delete volume {} ({})", node.hostname, params.name, vol.mountpoint_identifier.as_ref().unwrap());
    let pve_node = svc.pve_node(&node.pve_host)
        .map_err(P5xError::ServiceError)?;

    let mut pve_params = PutParams::default();
    pve_params.delete = vol.mountpoint_identifier;

    let vm_id = VmId::new(i64::from(node.pve_id)).unwrap();
    let r = pve_node
        .lxc()
        .vmid(vm_id)
        .config()
        .put(pve_params)
        .map_err(P5xError::PveError);

    if let Err(P5xError::PveError(proxmox_api::UreqError::Ureq(ureq::Error::Status(status, response)))) = r {
        let json_str = response.into_string().unwrap();
        error!("Error response from PVE API (status: {status}): {json_str}");
    }
    debug!("Successfully patched node {} config to delete volume {}", node.hostname, params.name);

    // Persist the volume
    volumes::Entity::delete_by_id(vol.volume_id)
        .exec(svc.db)
        .await
        .map_err(P5xError::DbErr)?;

    info!("Successfully deleted volume {}#{}", vol.name, vol.volume_id);
    Ok(())
}


/** Migrate a volume from its current LXC container to the specified LXC container. */
pub async fn transfer(
    svc: &Services<'_>,
    params: &VolumeParams,
    to_node: &nodes::Model,
) -> Result<volumes::Model, P5xError> {

    // Look up the volume from the params
    let vol = volumes::resolve(svc, params).await?;
    let from_node = vol.node(svc).await?;

    // If the volume already resides on to_node, we're done
    if from_node.pve_id == to_node.pve_id {
        return Ok(vol);
    }

    // If from_node and to_node are on the same physical host, transfer directly
    if from_node.pve_host == to_node.pve_host {
        return transfer_directly(svc, vol, &from_node, to_node).await;
    }

    // If the nodes are on different physical hosts, we need to create a temporary
    // container on shared storage to attach the volume to. We'll then migrate that
    // container to the target physical host.
    let carrier = provision_carrier(svc, &from_node).await?;
    let vol = transfer_directly(svc, vol, &from_node, &carrier).await?;

    let carrier = migrate_node(svc, carrier, &to_node.pve_host).await?;
    let vol = transfer_directly(svc, vol, &carrier, &to_node).await?;

    terminate_carrier(svc, carrier).await?;

    Ok(vol)
}


/** Migrate a volume from one LXC container to another LXC container, when both reside on the same PVE host. */
async fn transfer_directly(
    svc: &Services<'_>,
    vol: volumes::Model,
    from_node: &nodes::Model,
    to_node: &nodes::Model,
) -> Result<volumes::Model, P5xError> {
    if vol.mountpoint.is_some() {
        return Err(P5xError::BadPrecondition("Cannot transfer volume: volume is still mounted"));
    }

    if vol.mountpoint_identifier.is_none() {
        return Err(P5xError::BadPrecondition("Cannot transfer volume: volume is missing mountpoint identifier"));
    }

    if from_node.pve_host != to_node.pve_host {
        return Err(P5xError::BadPrecondition("Cannot transfer volume: nodes reside on different physical hosts"));
    }

    // Figure out the mountpoint identifier on the new node
    let _lock = to_node.lock(svc, Some("Receiving volume transfer"))
        .await.map_err(P5xError::LockErr)?;

    let config = to_node.config(svc)?;
    let mountpoint_identifier = config.next_nth("unused");
    let mountpoint_identifier = format!("unused{mountpoint_identifier}");

    // Ask the PVE API to move the volume to the new node
    let old_mountpoint_identifier = vol.mountpoint_identifier.as_ref().unwrap().as_str();
    let pve_from_vol = move_volume::Volume::try_from(old_mountpoint_identifier).unwrap();
    let pve_to_vol = move_volume::TargetVolume::try_from(mountpoint_identifier.as_str()).unwrap();
    let mut post_params = move_volume::PostParams::new(pve_from_vol);
    post_params.target_volume = Some(pve_to_vol);
    post_params.target_vmid = Some(to_node.vm_id());

    let res = svc.pve_node(&from_node.pve_host)
        .map_err(P5xError::ServiceError)?
        .lxc()
        .vmid(from_node.vm_id())
        .move_volume()
        .post(post_params);

    // This is necessary because POST returns {data: null} on success,
    // which the UreqClient flags as an unknown error.
    if let Err(UreqError::EncounteredErrors(e)) = res {
        return Err(P5xError::PveError(UreqError::EncounteredErrors(e)));
    }

    // Wait for the volume to transfer, since we don't get back a UPID
    sleep(Duration::from_secs(5)).await;

    // Verify that the volume appears in the new node's config
    let config = to_node.config(svc)?;
    let mount = config.get(&mountpoint_identifier)
        .ok_or(P5xError::BadPostcondition("Could not find mountpoint config after transferring volume"))?;

    // Parse the disk name from the config
    // synology-scsi-lun:vm-103-disk-1,mp=/mnt/p5x-test0,backup=1,size=1G
    let storage = &svc.config.pve_storage_pool;
    let name_offset = storage.len() + 1;  // 1 for the colon (:)
    let disk_name = mount[name_offset..].split(",").next().unwrap();
debug!("transfer_directly: mount {mount} | name_offset {name_offset} | disk_name {disk_name}");
    // Persist the volume
    let mut vol = vol.into_active_model();
    vol.pve_node_id = Set(to_node.pve_id);
    vol.mountpoint_identifier = Set(Some(mountpoint_identifier));
    vol.disk_name = Set(Some(disk_name.to_string()));

    let vol = vol.save(svc.db).await
        .map_err(P5xError::DbErr)?;

    volumes::Entity::find_by_id(vol.volume_id.unwrap())
        .one(svc.db)
        .await
        .map_err(P5xError::DbErr)?
        .ok_or(P5xError::BadPostcondition("Could not look up volume after persisting"))
}
