use std::time::Duration;
use rocket::futures;
use sea_orm::ActiveValue::Set;
use sea_orm::entity::prelude::*;
use sea_orm::{IntoActiveModel};
use serde::{Deserialize, Serialize};
use tokio::time::sleep;
use uuid::Uuid;
use log::info;

/** Try to acquire the given lock, if it is not held. */
pub async fn try_lock<'a>(
    db: &'a DatabaseConnection,
    lock_type: &str,
    lock_resource: &str,
    lock_reason: Option<&str>,
) -> Result<Option<LockHandle<'a>>, DbErr> {
    let owner = Uuid::new_v4();
    let mut l = ActiveModel::new();
    l.lock_owner = Set(Some(owner.to_string()));
    l.lock_reason = Set(lock_reason.map(|s| s.to_string()));

    Entity::update_many()
        .filter(Column::LockOwner.is_null())
        .filter(Column::LockType.eq(lock_type))
        .filter(Column::LockResource.eq(lock_resource))
        .set(l)
        .exec(db)
        .await?;

    Entity::find()
        .filter(Column::LockOwner.eq(owner.to_string()))
        .filter(Column::LockType.eq(lock_type))
        .filter(Column::LockResource.eq(lock_resource))
        .one(db)
        .await
        .map(|lock|
            lock.map(|lock | LockHandle::new(lock, db)))
}


/** Acquire the given lock, lazy-waiting until it is available. */
pub async fn lock<'a>(
    db: &'a DatabaseConnection,
    lock_type: &str,
    lock_resource: &str,
    lock_reason: Option<&str>,
) -> Result<LockHandle<'a>, DbErr> {
    loop {
        let r = try_lock(db, lock_type, lock_resource, lock_reason).await?;

        if let Some(r) = r {
            return Ok(r);
        }

        sleep(Duration::from_secs(10)).await;
    }
}


/** Acquire the global lock the overall cluster, lazy-waiting until it is available. (Used to create new LXC containers.) */
pub async fn lock_vmid<'a>(db: &'a DatabaseConnection, lock_reason: Option<&str>) -> Result<LockHandle<'a>, DbErr> {
    lock(db, "global_vmid", "0", lock_reason).await
}


/** Ensure the global VMID lock (used by lock_vmid) exists. */
pub async fn ensure_vmid_lock(db: &DatabaseConnection) -> Result<(), DbErr> {
    // Try to find the lock record
    let existing = Entity::find()
        .filter(Column::LockType.eq("global_vmid"))
        .filter(Column::LockResource.eq("0"))
        .one(db)
        .await?;

    if let Some(_) = existing {
        return Ok(());
    }

    // Create the lock record
    info!(target: "p5x", "Creating global VMID lock... (one-time fix up)");
    ActiveModel {
        lock_type: Set("global_vmid".to_string()),
        lock_resource: Set("0".to_string()),
        ..Default::default()
    }
        .save(db)
        .await?;

    Ok(())
}


/** A held lock. */
#[derive(Clone)]
pub struct LockHandle<'a> {
    pub lock: Model,
    db: &'a DatabaseConnection,
}

impl<'a> LockHandle<'a> {
    async fn release(&mut self) {
        debug!("Releasing {} {} lock", self.lock.lock_type, self.lock.lock_resource);

        let mut l = self.lock.clone().into_active_model();
        l.lock_owner = Set(None);
        l.lock_reason = Set(None);
        l.save(self.db).await.unwrap();
    }

    fn new(lock: Model, db: &'a DatabaseConnection) -> LockHandle<'a> {
        LockHandle { lock, db }
    }
}

impl<'a> Drop for LockHandle<'a> {
    fn drop(&mut self) {
        futures::executor::block_on(self.release());
    }
}

#[derive(Clone, Debug, PartialEq, Eq, DeriveEntityModel, Deserialize, Serialize, FromForm)]
#[sea_orm(table_name = "locks")]
pub struct Model {
    #[sea_orm(primary_key)]
    pub id: i32,
    pub lock_type: String,
    pub lock_resource: String,
    pub lock_owner: Option<String>,
    pub lock_reason: Option<String>,
}

#[derive(Copy, Clone, Debug, EnumIter, DeriveRelation)]
pub enum Relation {}

impl ActiveModelBehavior for ActiveModel {}
