use std::fmt::Display;
use std::net::TcpStream;
use ssh2::{OpenFlags, OpenType, Session};
use std::io::{Read, Write};
use std::path::Path;
use std::time::Duration;
use sea_orm::entity::prelude::*;
use sea_orm::QueryOrder;
use serde::{Deserialize, Serialize};
use tokio::time::sleep;
use proxmox_api;
use log::{warn, debug};
use proxmox_api::types::VmId;
use crate::api::entity::locks;
use crate::api::entity::locks::{lock, try_lock, LockHandle};
use crate::api::services::{Services, ServiceError, SshError, ssh_run_trimmed};

#[derive(Debug)]
pub enum NodeLockErr {
    DbErr(DbErr),
    WaitTimeLimitExceeded,
}

#[derive(Debug)]
pub enum P5xError {
    DbErr(DbErr),
    LockErr(NodeLockErr),
    ServiceError(ServiceError),
    PveError(proxmox_api::UreqError),
    SshError(SshError),
    InvalidNetworkInterface,
    BadPrecondition(&'static str),
    BadPostcondition(&'static str),
    UpidFailed(String, String),
}


/**
 * Acquire the first free LXC container lock.
 * Useful for operations that can occur on an arbitrary node (e.g. creating a new volume).
 * Lazy-wait until a lock is acquired.
 */
pub async fn lock_first_available<'a>(
    db: &'a DatabaseConnection,
    lock_reason: Option<&str>,
) -> Result<(Model, locks::LockHandle<'a>), NodeLockErr> {
    for _ in 0..600  {
        let r = try_lock_first_available(db, lock_reason)
            .await
            .map_err(NodeLockErr::DbErr)?;

        if let Some(r) = r {
            return Ok(r);
        }

        sleep(Duration::from_secs(10)).await;
    }

    warn!("Failed to acquire a lock on ANY nodes before the timeout was exceeded!");
    Err(NodeLockErr::WaitTimeLimitExceeded)
}


/**
 * Try to acquire the first free LXC container lock.
 * Useful for operations that can occur on an arbitrary node (e.g. creating a new volume).
 */
pub async fn try_lock_first_available<'a>(
    db: &'a DatabaseConnection,
    lock_reason: Option<&str>,
) -> Result<Option<(Model, locks::LockHandle<'a>)>, DbErr> {
    debug!("Trying to lock first available node. Reason: {}", lock_reason.unwrap_or("(none)"));

    // Get all permanent nodes
    let nodes = Entity::find()
        .filter(Column::IsPermanent.eq(true))
        .order_by_asc(Column::Id)
        .all(db)
        .await?;

    // Try to lock them and return the first successful one
    for node in nodes {
        let lock =
            try_lock(db, "nodes", &node.id.to_string(), lock_reason)
                .await?;

        if let Some(lock) = lock {
            info!("Locked node {} ({}): {}", node.hostname, node.pve_id, lock_reason.unwrap_or("(none)"));
            return Ok(Some((node, lock)))
        }
    }

    // No available nodes to lock
    Ok(None)
}


/** The PVE config file for an LXC container (i.e. /etc/pve/lxc/XYZ.conf) */
pub struct PveConfig {
    lines: Vec<String>,
}

impl PveConfig {
    pub fn new(conf: &str) -> PveConfig {
        let lines = conf.split("\n")
            .map(|s| s.to_string())
            .collect();

        PveConfig { lines }
    }

    /** Replace a line in the config file. */
    pub fn replace<F1, F2>(
        &mut self,
        matcher: F1,
        replacer: F2,
    )
    where
        F1: Fn(&String) -> bool,
        F2: Fn(&String) -> Option<String>
    {
        self.map(|s: &String| {
            if matcher(s) {
                return replacer(s);
            }

            Some(s.to_string())
        });
    }

    /** Run a map function on every line in the config file. */
    pub fn map<F>(
        &mut self,
        replacer: F
    )
    where
        F: FnMut(&String) -> Option<String>
    {
        self.lines = self.lines
            .iter()
            .filter_map(replacer)
            .collect();
    }

    /**
     * Get the config setting based on its key.
     * Example: If key="fubar", it will find the line "fubar: something"
     *          and return "something"
     */
    pub fn get(&self, key: &str) -> Option<String> {
        let key = format!("{key}:");
        let val = self.lines
            .iter()
            .filter(|line| line.starts_with(&key))
            .next()
            .map(|s| s.to_string())?;

        let offset = key.len() + 1;
        let val = val[offset..].trim();
        Some(val.to_string())
    }

    /**
     * Proxmox stores configs for resources that can have duplicates (e.g. volumes) in a predictable
     * format. For example, volumes will have "volume0: ..." / "volume1: ..." / &c.
     * This method will find the first free number for a given prefix.
     * Using that example, next_nth("volume") would return 2.
     */
    pub fn next_nth(&self, prefix: &str) -> u32 {
        let res = self.lines
            .iter()
            .filter(|line| line.starts_with(prefix))
            .map(|line|
                line[prefix.len()..]
                    .split(":")
                    .next()
                    .unwrap()
                    .parse::<u32>())
            .filter_map(|idx| idx.ok())
            .next();

        if let Some(idx) = res {
            return idx + 1;
        }

        0
    }
}

impl Display for PveConfig {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.lines.join("\n"))
    }
}


/** API-parameter version of a Node, used for form data I/O */
#[derive(Serialize, Deserialize, FromForm)]
pub struct NodeParams {
    pub id: Option<i32>,
    pub hostname: String,
    pub pve_id: i32,
    pub pve_host: String,
    pub assigned_ip: String,
    pub assigned_subnet: u32,
    pub is_permanent: bool,
}

impl NodeParams {
    /** Look up a node by hostname and get its params. */
    pub async fn resolve(svc: &Services<'_>, hostname: &str) -> Result<Option<NodeParams>, DbErr> {
        Entity::find()
            .filter(Column::Hostname.eq(hostname))
            .one(svc.db)
            .await
            .map(|v| v.map(|v| v.into()))
    }
}


#[derive(Clone, Debug, PartialEq, Eq, DeriveEntityModel, Deserialize, Serialize, FromForm)]
#[sea_orm(table_name = "nodes")]
pub struct Model {
    #[sea_orm(primary_key)]
    pub id: i32,
    pub hostname: String,
    pub pve_id: i32,
    pub pve_host: String,
    pub assigned_ip: String,
    pub assigned_subnet: u32,
    pub is_permanent: bool,
}

impl Model {
    /** Get the VM ID used by the Proxmox API library. */
    pub fn vm_id(&self) -> VmId {
        VmId::new(i64::from(self.pve_id)).unwrap()
    }

    /** Open an SSH session to this LXC container. */
    pub fn ssh(&self, svc: &Services) -> Result<Session, P5xError> {
        let addr = format!("{}:22", self.assigned_ip);
        let tcp = TcpStream::connect(addr)
            .map_err(SshError::IoError)
            .map_err(P5xError::SshError)?;

        let mut sess = Session::new()
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        sess.set_tcp_stream(tcp);
        sess.handshake()
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        let pubkey = &svc.config.ssh_pubkey;
        let privkey = &svc.config.ssh_privkey;

        sess.userauth_pubkey_memory("root", Some(pubkey), privkey, None)
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        if !sess.authenticated() {
            return Err(P5xError::SshError(SshError::AuthFailed));
        }

        Ok(sess)
    }

    /** Load the LXC container config for this node. */
    pub fn config(&self, svc: &Services) -> Result<PveConfig, P5xError> {
        let pve_ssh = svc.pve_ssh(&self.pve_host)
            .map_err(P5xError::ServiceError)?;

        let path = format!("/etc/pve/lxc/{}.conf", self.pve_id);
        let path = Path::new(&path);

        let (mut remote_file, _) = pve_ssh.scp_recv(path)
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        let mut contents = String::new();
        remote_file.read_to_string(&mut contents)
            .map_err(SshError::IoError)
            .map_err(P5xError::SshError)?;

        // Close the channel and wait for the whole content to be transferred
        remote_file.send_eof().map_err(SshError::ClientError).map_err(P5xError::SshError)?;
        remote_file.wait_eof().map_err(SshError::ClientError).map_err(P5xError::SshError)?;
        remote_file.close().map_err(SshError::ClientError).map_err(P5xError::SshError)?;
        remote_file.wait_close().map_err(SshError::ClientError).map_err(P5xError::SshError)?;

        Ok(PveConfig::new(&contents))
    }

    /** Replace the LXC container config for this node. */
    pub fn write_config(&self, svc: &Services, conf: &PveConfig) -> Result<(), P5xError> {
        let conf = conf.to_string();
        let conf = conf.as_bytes();
        let pve_ssh = svc.pve_ssh(&self.pve_host)
            .map_err(P5xError::ServiceError)?;

        // Write the file
        let sftp = pve_ssh.sftp()
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        let path = format!("/etc/pve/lxc/{}.conf", self.pve_id);
        let path = Path::new(&path);

        debug!("Attempting to open file to mutate config {} on {}", path.display(), self.pve_host);
        let mut f = sftp.open_mode(path, OpenFlags::WRITE, 0o640, OpenType::File)
            .map_err(SshError::ClientError)
            .map_err(P5xError::SshError)?;

        debug!("Attempting to write config to {} on {}", path.display(), self.pve_host);
        f.write_all(conf)
            .map_err(SshError::IoError)
            .map_err(P5xError::SshError)?;

        Ok(())
    }

    /** Try to acquire the node lock for this LXC container. Used for any operations that impact the config. */
    pub async fn try_lock<'a>(&self, svc: &Services<'a>, lock_reason: Option<&str>) -> Result<Option<LockHandle<'a>>, NodeLockErr> {
        try_lock(svc.db, "nodes", &self.id.to_string(), lock_reason)
            .await
            .map_err(NodeLockErr::DbErr)
    }

    /** Block until we acquire the node lock for this LXC container. Used for any operations that impact the config. */
    pub async fn lock<'a>(&self, svc: &Services<'a>, lock_reason: Option<&str>) -> Result<LockHandle<'a>, NodeLockErr> {
        lock(svc.db, "nodes", &self.id.to_string(), lock_reason)
            .await
            .map_err(NodeLockErr::DbErr)
    }

    /** Helper for loading, modifying, then replacing the config, while managing the lock automatically. */
    pub async fn mutate_config<F>(
        &self,
        svc: &Services<'_>,
        lock_reason: Option<&str>,
        mutator: F,
    ) -> Result<(), P5xError>
    where
        F: FnOnce(PveConfig) -> PveConfig
    {
        debug!("Mutating config on node {}", self.hostname);
        let _lock = self.lock(svc, lock_reason)
            .await.map_err(P5xError::LockErr)?;

        let config = self.config(svc)?;
        debug!("Successfully loaded config for mutate on node {}", self.hostname);
        let config = mutator(config);

        self.write_config(svc, &config)
    }

    /** Run an SSH command on the node and return the output of the command, with whitespace trimmed. */
    pub fn ssh_run_trimmed(&self, svc: &Services, cmd: &str) -> Result<String, P5xError> {
        let node_ssh = self.ssh(svc)?;
        ssh_run_trimmed(&node_ssh, cmd)
    }
}

#[derive(Copy, Clone, Debug, EnumIter, DeriveRelation)]
pub enum Relation {}

impl ActiveModelBehavior for ActiveModel {}

impl Into<NodeParams> for Model {
    fn into(self) -> NodeParams {
        NodeParams {
            id: Some(self.id),
            hostname: self.hostname,
            pve_id: self.pve_id,
            pve_host: self.pve_host,
            assigned_ip: self.assigned_ip,
            assigned_subnet: self.assigned_subnet,
            is_permanent: self.is_permanent,
        }
    }
}
