use std::io::Read;
use std::net::TcpStream;
use proxmox_api::nodes::node::NodeClient;
use proxmox_api::UreqClient;
use sea_orm::{DatabaseConnection, DbErr};
use ssh2::Session;
use crate::api::entity::nodes::P5xError;
use crate::api::util::{read_p5x_config, P5xConfig};

#[derive(Debug)]
pub enum SshError {
    IoError(std::io::Error),
    ClientError(ssh2::Error),
    CommandFailed(i32, String),
    AuthFailed,
}

impl SshError {
    pub fn command_failed(exit_code: i32, command: &str, output: &str) -> SshError {
        SshError::CommandFailed(exit_code, format!("Command `{command}` failed: {output}"))
    }
}

#[derive(Debug)]
pub enum ServiceError {
    PveError(proxmox_api::UreqError),
    InvalidNetworkInterface,
    SshError(SshError),
}


/** Helper for managing singleton services. */
pub struct Services<'a> {
    pub db: &'a DatabaseConnection,
    pub config: P5xConfig,
}

impl<'a> Services<'a> {
    /** Create a new services instance. */
    pub async fn build(db: &'a DatabaseConnection) -> Result<Services<'a>, DbErr> {
        Ok(Services {
            db,
            config: read_p5x_config(),
        })
    }

    /** Get a new Proxmox API client instance. */
    pub fn pve(&self) -> Result<UreqClient, ServiceError> {
        let host = &self.config.pve_api_host;
        let pw = &self.config.pve_root_password;

        let host = format!("https://{host}:8006");
        let api = UreqClient::new(&host, "root", "pam", &pw)
            .map_err(ServiceError::PveError)?;

        Ok(api)
    }

    /** Get a Proxmox Node API client instance for the given node. */
    pub fn pve_node(&self, host: &str) -> Result<NodeClient<UreqClient>, ServiceError> {
        let pve = self.pve()?;
        Ok(proxmox_api::nodes::NodesClient::new(pve).node(host))
    }

    /** Given a PVE host name, get its network address. */
    pub fn pve_addr(&self, host: &str) -> Result<String, ServiceError> {
        let ifaces = self.pve_node(host)?
            .network()
            .get(Default::default())
            .map_err(ServiceError::PveError)?;

        let addr = ifaces.iter()
            .filter_map(|i| i.additional_properties.get("address"))
            .next()
            .ok_or(ServiceError::InvalidNetworkInterface)?;

        if !addr.is_string() {
            return Err(ServiceError::InvalidNetworkInterface);
        }

        Ok(addr.as_str().unwrap().to_string())
    }

    /** Open a new SSH session to the given PVE node. */
    pub fn pve_ssh(&self, host: &str) -> Result<Session, ServiceError> {
        let addr = self.pve_addr(host)?;
        let addr = format!("{}:22", addr);
        let tcp = TcpStream::connect(addr)
            .map_err(SshError::IoError)
            .map_err(ServiceError::SshError)?;

        let mut sess = Session::new()
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        sess.set_tcp_stream(tcp);
        sess.handshake()
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        sess.userauth_password("root", &self.config.pve_root_password)
            .map_err(SshError::ClientError)
            .map_err(ServiceError::SshError)?;

        if !sess.authenticated() {
            return Err(ServiceError::SshError(SshError::AuthFailed));
        }

        Ok(sess)
    }
}


/** Run an SSH command and return the output as a string, with whitespace trimmed. */
pub fn ssh_run_trimmed(session: &Session, cmd: &str) -> Result<String, P5xError> {
    let mut channel = session.channel_session()
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    channel.exec(cmd)
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    let mut contents = String::new();
    channel.read_to_string(&mut contents)
        .map_err(SshError::IoError)
        .map_err(P5xError::SshError)?;

    let exit_code = channel.exit_status()
        .map_err(SshError::ClientError)
        .map_err(P5xError::SshError)?;

    let contents = contents.trim();
    if exit_code != 0 {
        return Err(P5xError::SshError(SshError::command_failed(exit_code, cmd, &contents)));
    }

    Ok(contents.to_string())
}
